/*
 * Copyright (C) 2020 Activision Publishing, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * 3. Neither the name of Activision Publishing, Inc. nor the names of
 *    its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "Universal/TelescopePlatformConfig.h"
#include "TelescopeDefaultAllocator.h"
#include "TelescopeMemoryTracing.h"

#if TELESCOPE( PC_PROGRAM )  || TELESCOPE( XB_PROGRAM )
#include <windows.h>
#endif

namespace Telescope
{
#if TELESCOPE( PC_PROGRAM ) || TELESCOPE( XB_PROGRAM )
static inline DWORD protection( bool writable, bool executable )
{
	return executable ?
		( writable ? PAGE_EXECUTE_READWRITE : PAGE_EXECUTE_READ ) :
		( writable ? PAGE_READWRITE : PAGE_READONLY );
}
#endif


void* DefaultMemoryAllocator::Allocate( size_t size )
{
	void* result = malloc(size);
	if( !result )
		CRASH();

#if TELESCOPE( MEMORY_TRACING )
	TelescopeMemoryTracing::GetInstance()->CheckBacktrace( result );
#endif

	return result;
}


void* DefaultMemoryAllocator::AllocateAligned( size_t size, size_t align )
{
#if TELESCOPE( PC_PROGRAM ) || TELESCOPE( XB_PROGRAM )
	void* result = _aligned_malloc( size, align );
#elif TELESCOPE( SONY_PROGRAM )
	void* result = memalign( align, size );
#endif
	if( !result )
		CRASH();
	return result;
}


void* DefaultMemoryAllocator::Reallocate( void* p, size_t size )
{
	void* result = realloc( p, size );
	if( !result )
		CRASH();
	return result;
}


void DefaultMemoryAllocator::Deallocate( void* p, size_t size )
{
	free( p );

#if TELESCOPE( MEMORY_TRACING )
	TelescopeMemoryTracing::GetInstance()->RemoveBacktrace( p );
#endif
}


void DefaultMemoryAllocator::DeallocateAligned( void* p, size_t size )
{
#if TELESCOPE( PC_PROGRAM ) || TELESCOPE( XB_PROGRAM )
	_aligned_free( p );
#elif TELESCOPE( SONY_PROGRAM )
	free( p );
#endif
}


size_t DefaultMemoryAllocator::AllocateSize( void* p )
{
#if TELESCOPE( PC_PROGRAM ) || TELESCOPE( XB_PROGRAM )
	return _msize( p );
#elif TELESCOPE( SONY_PROGRAM )	
	return malloc_usable_size( p );
#endif
}


void* DefaultMemoryAllocator::ReserveUncommitted( size_t size, bool writable, bool executable )
{
#if TELESCOPE( PC_PROGRAM ) || TELESCOPE( XB_PROGRAM )
	void* result = VirtualAlloc( 0, size, MEM_RESERVE, protection( writable, executable ) );
#elif TELESCOPE( SONY_PROGRAM )
	// TODO
#ifdef NDEBUG
	void* result = nullptr;
#else
	void* result = malloc(size);
#endif
#endif
	if( !result )
		CRASH();
	return result;
}


void* DefaultMemoryAllocator::ReserveAndCommit( size_t size, bool writable, bool executable )
{
#if TELESCOPE( PC_PROGRAM ) || TELESCOPE( XB_PROGRAM )
	void* result = VirtualAlloc( 0, size, MEM_RESERVE | MEM_COMMIT, protection( writable, executable ) );
#elif TELESCOPE( SONY_PROGRAM )
	// TODO
#ifdef NDEBUG
	void* result = nullptr;
#else
	void* result = malloc( size );
#endif
#endif
	if( !result )
		CRASH();
	return result;
}


void* DefaultMemoryAllocator::Commit( void* address, size_t size, bool writable, bool execuable )
{
#if TELESCOPE( PC_PROGRAM ) || TELESCOPE( XB_PROGRAM )
	void* result = VirtualAlloc(address, size, MEM_COMMIT, protection( writable, execuable ) );
#elif TELESCOPE( SONY_PROGRAM )
	// TODO
#ifdef NDEBUG
	void* result = nullptr;
#else
	void* result = malloc( size );
#endif
#endif
	if ( !result )
		CRASH();
	return result;
}


void DefaultMemoryAllocator::Decommit( void* address, size_t size )
{
	// According to http://msdn.microsoft.com/en-us/library/aa366892(VS.85).aspx,
	// bytes (i.e. dwSize) being 0 when dwFreeType is MEM_DECOMMIT means that we'll
	// decommit the entire region allocated by VirtualAlloc() instead of decommitting
	// nothing as we would expect. Hence, we should check if bytes is 0 and handle it
	// appropriately before calling VirtualFree().
	// See: https://bugs.webkit.org/show_bug.cgi?id=121972.
	if( !size )
	{
		return;
	}
	// Silence warning about using MEM_DECOMMIT instead of MEM_RELEASE:

#if TELESCOPE( PC_PROGRAM ) || TELESCOPE( XB_PROGRAM )
#pragma warning(suppress: 6250)
	bool result = VirtualFree( address, size, MEM_DECOMMIT );
#elif TELESCOPE( SONY_PROGRAM )
	bool result = false;
#endif

	if( !result )
		CRASH();
}


void DefaultMemoryAllocator::ReleaseDecommit( void* address, size_t size )
{
	// See comment in DefaultMemoryAllocator::Decommit(). Similarly, when bytes is 0, we
	// don't want to release anything. So, don't call VirtualFree() below.
	if( !size )
	{
		return;
	}
	// According to http://msdn.microsoft.com/en-us/library/aa366892(VS.85).aspx,
	// dwSize must be 0 if dwFreeType is MEM_RELEASE.
#if TELESCOPE( PC_PROGRAM ) || TELESCOPE( XB_PROGRAM )
	bool result = VirtualFree( address, 0, MEM_RELEASE );
#elif TELESCOPE( SONY_PROGRAM )
	bool result = false;
#endif

	if( !result )
		CRASH();
}
}
