/*
 * Copyright (C) 2020 Activision Publishing, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * 3. Neither the name of Activision Publishing, Inc. nor the names of
 *    its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "Universal/TelescopePlatformConfig.h"
#include "Universal/TelescopeLog.h"
#include "TelescopeLib.h"
#include "WebView.h"
#include "ClientWrappers/MemoryAllocatorBase.h"
#include "ClientWrappers/CURLReplacement/CurlReplacementImpl.h"
#include "ClientWrappers/HashingReplacement/HashingReplacement.h"
#include "ClientWrappers/ICUMMAPReplacement/IcuMmpReplacement.h"
#include "ClientWrappers/SystemAPIReplacement/SystemAPIReplacement.h"
#include "ClientWrappers/ErrorHandlingReplacement/ErrorHandlingReplacement.h"
#include "ClientWrappers/LogReplacement/LogReplacement.h"
#include "wtf/MainThread.h"
#include "wtf/RunLoop.h"
#include "wtf/WTFLog2.h"
#include "wtf/TelescopePrintStream.h"
#include "fontconfig/fontconfig.h"
#include "MainThreadSharedTimer.h"
#include "Logging.h"
#if TELESCOPE(COMMAND_SYSTEM)
#include "TelescopeCMD.h"
#endif
#include <wtf/ThreadSpecific.h>
#include "network/curl/CurlShutdown.h"
#include <WebCore/CurlCacheManager.h>

extern void setCompressionWrapper(CompressionWrapperBase* pCompressionWrapper );
extern CompressionWrapperBase* getCompressionWrapper();
extern unsigned int g_uiTelescopeCL;

#if IS_DEBUG
#define sPublishVersion "Debug"
#elif IS_RELEASE
#define sPublishVersion "Release"
#elif IS_SHIP
#define sPublishVersion "Ship"
#else
#define sPublishVersion "Unknown"
#endif

namespace Telescope
{
#if TELESCOPE(COMMAND_SYSTEM)
TelescopeCMD* g_pTelescopeCMD = nullptr;
#endif

void TelescopeLib::PreInit( TelescopeConfig* config )
{
	Telescope::Telescope_InstallMemoryAllocator( config->memAllocator );
	Telescope::InstallErrorHandlingWrapper(config->errorHandlingImpl);
}

bool TelescopeLib::Init(TelescopeConfig* config )
{
	// TODO: the naming conversion are not identical to Telescope_InstallMemoryAllocator, fix it.
	if ( config->httpWrapper )
		Telescope::installHttpWrapper( config->httpWrapper );
	else
		ASSERT_WITH_MESSAGE( 0, "Telescope must have the implementation of HTTP." );

	if( config->hashingImpl )
		Telescope::installHashingWrapper( config->hashingImpl );
	else
		ASSERT_WITH_MESSAGE( 0, "Telescope must have the implementation of HASHING." );

	if ( config->compressionImpl )
		setCompressionWrapper( config->compressionImpl );
	else
		ASSERT_WITH_MESSAGE( 0, "Telescope must have the implementation of a Compression implementation." );

	if ( config->assetLoaderImpl )
		Telescope::installOpenFileWrapper( config->assetLoaderImpl );
	else
		ASSERT_WITH_MESSAGE( 0, "Telescope must have the implementation to asset loader." );

	if ( config->sysAPIImpl )
		Telescope::installSystemAPIWrapper( config->sysAPIImpl );
	else
		ASSERT_WITH_MESSAGE( 0, "Telescope must have the implementation of system API." );

	if ( config->logImpl )
		Telescope::InstallLogWrapper( config->logImpl );
	else
		ASSERT_WITH_MESSAGE( 0, "Telescope must have the implementation of log system." );

	WTF::RunLoop::initializeMainRunLoop();

#if TELESCOPE(COMMAND_SYSTEM)
	g_pTelescopeCMD = new TelescopeCMD();
#endif

	RELEASE_LOG_INFO( Telescope, "PublishVersion:[%s], Compile Time:[%s %s], CL: %u\n",
		sPublishVersion, __DATE__, __TIME__, g_uiTelescopeCL );

	return true;
}

void TelescopeLib::Shutdown()
{
	// TODO: Looks like Telescope is not shutdown when reaches here.
	// Telescope::Telescope_UninstallMemoryAllocator();
#if TELESCOPE(COMMAND_SYSTEM)
	if ( g_pTelescopeCMD )
	{
		delete g_pTelescopeCMD;
		g_pTelescopeCMD = nullptr;
	}
#endif

	WTF::RunLoop::shutdown();

#if USE(PTHREADS)
	WTF::shutdownPThreadKeyHashMap();
#endif

	WebCore::curlShutdown();
}

void TelescopeLib::Tick()
{
	WTF::dispatchFunctionsFromMainThread();

	WebCore::tickTimer();
}

WebView* TelescopeLib::CreateView()
{
	return new WebView();
}

void TelescopeLib::DestroyView(WebView* view)
{
	view->ShutdownView();
	delete view;
}

bool TelescopeLib::InstallFont(const unsigned char* fontData, size_t dataSize)
{
	return FcAddFontMemory(fontData, dataSize);
}

void TelescopeLib::InitFileCache(const char* cacheDirectory, size_t storageSizeLimitInBytes)
{
	String cacheFolder(cacheDirectory);
	WebCore::CurlCacheManager::singleton().setCacheDirectory(cacheFolder);
	WebCore::CurlCacheManager::singleton().setStorageSizeLimit(storageSizeLimitInBytes);
}

void TelescopeLib::ExecuteCMD( const char* pCmd, const char* pParam )
{
#if TELESCOPE(COMMAND_SYSTEM)
	if ( g_pTelescopeCMD )
		g_pTelescopeCMD->ExecuteCMD( pCmd, pParam );
#endif
}

unsigned int TelescopeLib::getTelescopeCL()
{
	return g_uiTelescopeCL;
}

void TelescopeLib::LogTypeKillSwitch( LogType type, bool enabled )
{
	Telescope::SwitchTelescopeLogType( type, enabled );
	WTF::SwitchMiscLogLevel( static_cast<unsigned int>(WTFLogLevel::Debug), enabled );
	if ( type == LOG_TYPE_INFO )
	{
		SwitchWTFLog2Level( static_cast<unsigned int>(WTFLogLevel::Always), enabled );
		SwitchWTFLog2Level( static_cast<unsigned int>(WTFLogLevel::Info), enabled );
	}
	else if ( type == LOG_TYPE_WARNING )
		SwitchWTFLog2Level( static_cast<unsigned int>(WTFLogLevel::Warning), enabled );
	else if ( type == LOG_TYPE_ERROR )
		SwitchWTFLog2Level( static_cast<unsigned int>(WTFLogLevel::Error), enabled );
	else if ( type == LOG_TYPE_DEBUG )
		SwitchWTFLog2Level( static_cast<unsigned int>(WTFLogLevel::Debug), enabled );
}

void TelescopeLib::LogChannelKillSwitch( const char* moduleName, const char* systemName, bool enabled )
{
#if !LOG_DISABLED || !RELEASE_LOG_DISABLED
	if ( equalIgnoringASCIICase( moduleName, "WEBCORE" ) )
		WebCore::switchWebCoreChannel( systemName, enabled );
	else
#endif
	if ( equalIgnoringASCIICase( moduleName, Telescope::LOG_MODULE_NAME ) )
		Telescope::SwitchTelescopeChannel( systemName, enabled );
	else if ( equalIgnoringASCIICase( moduleName, "MISC") )
		WTF::SwitchMiscLogChannel( enabled );
}

}
